// CoreCalc, a spreadsheet core implementation   

// ----------------------------------------------------------------------
// Copyright (c) 2006-2010 Peter Sestoft and others

// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

//  * The above copyright notice and this permission notice shall be
//    included in all copies or substantial portions of the Software.

//  * The software is provided "as is", without warranty of any kind,
//    express or implied, including but not limited to the warranties of
//    merchantability, fitness for a particular purpose and
//    noninfringement.  In no event shall the authors or copyright
//    holders be liable for any claim, damages or other liability,
//    whether in an action of contract, tort or otherwise, arising from,
//    out of or in connection with the software or the use or other
//    dealings in the software.
// ----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;

// Classes for representing absolute cell addresses,
// relative/absolute cell references, and 
// adjusted expressions and refererences

namespace CoreCalc {

  // ----------------------------------------------------------------
  // A CellAddr is an absolute, zero-based (col, row) location

  [Serializable]
  public struct CellAddr : IEquatable<CellAddr> {
    public readonly int col, row;
    public static readonly CellAddr A1 = new CellAddr(0, 0);

    public CellAddr(int col, int row) {
      this.col = col;
      this.row = row;
    }

    public CellAddr(RARef cr, int col, int row) {
      this.col = cr.colAbs ? cr.colRef : cr.colRef + col;
      this.row = cr.rowAbs ? cr.rowRef : cr.rowRef + row;
    }

    // Turn an A1-format string into an absolute cell address
    public CellAddr(String a1Ref) : this(new RARef(a1Ref, 0, 0), 0, 0) { }

    // Return ulCa as upper left and lrCa as lower right of (ca1, ca2)
    public static void NormalizeArea(CellAddr ca1, CellAddr ca2,
                                     out CellAddr ulCa, out CellAddr lrCa) {
      int minCol = ca1.col, minRow = ca1.row,
          maxCol = ca2.col, maxRow = ca2.row;
      if (ca1.col > ca2.col) {
        minCol = ca2.col; maxCol = ca1.col;
      }
      if (ca1.row > ca2.row) {
        minRow = ca2.row; maxRow = ca1.row;
      }
      ulCa = new CellAddr(minCol, minRow);
      lrCa = new CellAddr(maxCol, maxRow);
    }

    public CellAddr(System.Drawing.Point p) {
      this.col = p.X;
      this.row = p.Y;
    }

    public CellAddr Offset(CellAddr offset)
    {
      return new CellAddr(this.col + offset.col, this.row + offset.row);
    }

    public bool Equals(CellAddr that)
    {
      return col == that.col && row == that.row;
    }

    public override int GetHashCode()
    {
      return 27 * col + row;
    }

    public bool Equals(Object o)
    {
      return o is CellAddr && Equals((CellAddr)o);
    }

    public override String ToString() {
      return ColumnName(col) + (row + 1);
    }

    public static String ColumnName(int col) {
      String name = "";
      while (col >= 26) {
        name = (char)('A' + col % 26) + name;
        col = col / 26 - 1;
      }
      return (char)('A' + col) + name;
    }
  }

  // ----------------------------------------------------------------
  // A RARef represents a relative or absolute cell reference

  public sealed class RARef {
    public readonly bool colAbs, rowAbs;  // True=absolute, False=relative
    public readonly int colRef, rowRef;

    public RARef(bool colAbs, int colRef, bool rowAbs, int rowRef) {
      this.colAbs = colAbs; this.colRef = colRef;
      this.rowAbs = rowAbs; this.rowRef = rowRef;
    }

    // Parse "$A$3" to (true,0,true,2) and so on; relative references
    // must be adjusted to the (col,row) in which the reference occurs.

    public RARef(String a1Ref, int col, int row) {
      /* RC denotes "this" cell in MS XMLSS format; needed for area support */
      if (a1Ref.Equals("rc", StringComparison.CurrentCultureIgnoreCase)) {
        colAbs = false; rowAbs = false; colRef = 0; rowRef = 0;
      } else {
        int i = 0;
        if (i < a1Ref.Length && a1Ref[i] == '$') {
          colAbs = true;
          i++;
        }
        int val = -1;
        while (i < a1Ref.Length && IsAToZ(a1Ref[i])) {
          val = (val + 1) * 26 + AToZValue(a1Ref[i]);
          i++;
        }
        colRef = colAbs ? val : val - col;
        if (i < a1Ref.Length && a1Ref[i] == '$') {
          rowAbs = true;
          i++;
        }
        val = ParseInt(a1Ref, ref i);
        rowRef = (rowAbs ? val : val - row) - 1;
      }
    }

    // Parse XMLSS R1C1 notation from well-formed s (checked by parser)

    public RARef(String r1c1) {
      int i = 0;
      rowAbs = true;
      colAbs = true;
      if (i < r1c1.Length && r1c1[i] == 'R')
        i++;
      if (i < r1c1.Length && r1c1[i] == '[') {
        rowAbs = false;
        i++;
      }
      int val = ParseInt(r1c1, ref i);
      if (rowAbs && val == 0)
        rowAbs = false;
      this.rowRef = rowAbs ? val - 1 : val;
      if (i < r1c1.Length && r1c1[i] == ']')
        i++;
      if (i < r1c1.Length && r1c1[i] == 'C')
        i++;
      if (i < r1c1.Length && r1c1[i] == '[') {
        colAbs = false;
        i++;
      }
      val = ParseInt(r1c1, ref i);
      if (i < r1c1.Length && r1c1[i] == ']')
        i++;
      if (colAbs && val == 0)
        colAbs = false;
      this.colRef = colAbs ? val - 1 : val;
      Formats fo = new Formats();
      fo.RefFmt = Formats.RefType.R1C1;
    }

    // Parse possibly signed decimal integer
    private static int ParseInt(String s, ref int i) {
      int val = 0;
      bool negative = false;
      if (i < s.Length && (s[i] == '-' || s[i] == '+')) {
        negative = s[i] == '-';
        i++;
      }
      val = 0;
      while (i < s.Length && Char.IsDigit(s[i])) {
        val = val * 10 + (s[i] - '0');
        i++;
      }
      return negative ? -val : val;
    }

    private static bool IsAToZ(char c) {
      return 'a' <= c && c <= 'z' || 'A' <= c && c <= 'Z';
    }

    private static int AToZValue(char c) {
      return (c - 'A') % 32;
    }

    // Absolute address of ref 

    public CellAddr Addr(int col, int row) {
      return new CellAddr(this, col, row);
    }

    // Insert N new rowcols before rowcol R>=0, when we're at rowcol r
    public Adjusted<RARef> InsertRowCols(int R, int N, int r, bool insertRow) {
      int newRef;
      int upper;
      if (insertRow) {   // Insert N rows before row R; we're at row r
        InsertRowCols(R, N, r, rowAbs, rowRef, out newRef, out upper);
        RARef rarefNew = new RARef(colAbs, colRef, rowAbs, newRef);
        return new Adjusted<RARef>(rarefNew, upper, rowRef == newRef);
      } else {           // Insert N columns before column R; we're at column r
        InsertRowCols(R, N, r, colAbs, colRef, out newRef, out upper);
        RARef rarefNew = new RARef(colAbs, newRef, rowAbs, rowRef);
        return new Adjusted<RARef>(rarefNew, upper, colRef == newRef);
      }
    }

    // Insert N new rowcols before rowcol R>=0, when we're at rowcol r
    private static void InsertRowCols(int R, int N, int r,
                                      bool rcAbs, int rcRef,
                                      out int newRc, out int upper) 
    {
      if (rcAbs) {
        if (rcRef >= R) {
          // Absolute ref to cell after inserted          // Case (Ab)
          newRc = rcRef + N;
          upper = int.MaxValue;
        } else {
          // Absolute ref to cell before inserted       // Case (Aa) 
          newRc = rcRef;
          upper = int.MaxValue;
        }
      } else // Relative reference
        if (r >= R) {
          if (r + rcRef < R) {
            // Relative ref from after inserted rowcols to cell before them
            newRc = rcRef - N;                          // Case (Rab)
            upper = R - rcRef;
          } else {
            // Relative ref from after inserted rowcols to cell after them
            newRc = rcRef;                              // Case (Rbb)
            upper = int.MaxValue;
          }
        } else // r < R
          if (r + rcRef >= R) {
            // Relative ref from before inserted rowcols to cell after them
            newRc = rcRef + N;                          // Case (Rba)
            upper = R;
          } else {
            // Relative ref from before inserted rowcols to cell before them
            newRc = rcRef;                              // Case (Raa)
            upper = Math.Min(R, R - rcRef);
          }
    }

    // Clone and move (when the containing formula is moved, not copied)

    public RARef Move(int deltaCol, int deltaRow) {
      return new RARef(colAbs, colAbs ? colRef : colRef + deltaCol,
                       rowAbs, rowAbs ? rowRef : rowRef + deltaRow);
    }

    public String Show(int col, int row, Formats fo) {
      switch (fo.RefFmt) {
        case Formats.RefType.A1:
          CellAddr ca = new CellAddr(this, col, row);
          return (colAbs ? "$" : "") + CellAddr.ColumnName(ca.col)
               + (rowAbs ? "$" : "") + (ca.row + 1);
        case Formats.RefType.R1C1:
          return "R" + RelAbsFormat(rowAbs, rowRef, 1)
               + "C" + RelAbsFormat(colAbs, colRef, 1);
        case Formats.RefType.C0R0:
          return "C" + RelAbsFormat(colAbs, colRef, 0)
               + "R" + RelAbsFormat(rowAbs, rowRef, 0);
        default:
          throw new ImpossibleException("Unknown reference format");
      }
    }

    private static String RelAbsFormat(bool abs, int offset, int origo) {
      if (abs)
        return (offset + origo).ToString();
      else if (offset == 0)
        return "";
      else
        return "[" + (offset > 0 ? "+" : "") + offset.ToString() + "]";
    }
  }

  // ----------------------------------------------------------------
  // An Adjusted<T> represents an adjusted expression or 
  // a relative/absolute ref, for InsertRowCols

  public struct Adjusted<T> {
    public readonly T e;          // The adjusted Expr or RaRef
    public readonly int upper;    // ... invalid for rows >= upper
    public readonly bool same;    // Adjusted is identical to original

    public Adjusted(T e, int upper, bool same) {
      this.e = e;
      this.upper = upper;
      this.same = same;
    }

    public Adjusted(T e) : this(e, int.MaxValue, true) { }
  }
}
